# `xml-disassembler`

[![NPM](https://img.shields.io/npm/v/xml-disassembler.svg?label=xml-disassembler)](https://www.npmjs.com/package/xml-disassembler) [![Downloads/week](https://img.shields.io/npm/dw/xml-disassembler.svg)](https://npmjs.org/package/xml-disassembler)

Disassemble XML files into smaller, more manageable files and reassemble them when needed.

This tool simplifies version control, improves diff readability, and streamlines collaboration when dealing with large XML files.

---

## 🚀 Features

- **Disassemble XML Files** – Break down XML files into structured directories.
- **Reassemble XML Files** – Recreate the original XML structure from disassembled parts.
- **Unique Identifiers** – Use specific XML elements as file names or fallback to SHA-256 hashes.
- **Ignore Files** – Specify XML files to exclude from disassembly.
- **Logging** – Enable detailed debugging logs.
- **Integrations** – Works with tools like Salesforce CLI
- **Extensions** - Extensions allow for JSON, YAML, and JSON5 transformations.

<!-- TABLE OF CONTENTS -->
<details>
  <summary>Table of Contents</summary>

- [Background](#background)
- [Trade-offs](#trade-offs)
- [Install](#install)
- [Disassembling Files](#disassembling-files)
- [Reassembling Files](#reassembling-files)
- [Example](#example)
- [Use Case](#use-case)
- [Ignore File](#ignore-file)
- [XML Parser](#xml-parser)
- [Logging](#logging)
- [Extensions](#extensions)
- [Contributing](#contributing)
- [Template](#template)
</details>

## Background

Large XML files, especially those generated by external tools, can be challenging to review and manage.  
`xml-disassembler` helps by breaking down these files into smaller, more digestible chunks, making it easier to track changes and collaborate.

## Trade-offs

Instead of relying on complex XML diff algorithms, `xml-disassembler` provides a simple and accessible solution by splitting XML files into structured directories.

## Install

Install the package using NPM:

```
npm install xml-disassembler
```

## Disassembling Files

Disassemble a single XML file or multiple XML files within a directory.

```typescript
/* 
FLAGS
- filePath:         Relative path to 1 XML file or a directory of XML files to disassemble.
- uniqueIdElements: Comma-separated list of unique and required ID elements used to name disassembled files for nested elements. 
                    Defaults to SHA-256 hash if unique ID elements are undefined or not found.
- prePurge:         Delete pre-existing disassembled directories prior to disassembling the file.
- postPurge:        Delete the original XML file after disassembling it.

- ignorePath:       Path to an disassembly ignore file.
*/
import { DisassembleXMLFileHandler } from "xml-disassembler";

const handler = new DisassembleXMLFileHandler();
await handler.disassemble({
  filePath: "test/baselines/general",
  uniqueIdElements:
    "application,apexClass,name,externalDataSource,flow,object,apexPage,recordType,tab,field",
  prePurge: true,
  postPurge: true,
  ignorePath: ".xmldisassemblerignore",
});
```

## Reassembling Files

Reassemble a directory of disassembled XML files into a single XML file.

```typescript
/* 
FLAGS
- filePath:        Relative path to the disassembled XML directory to reassemble.
- fileExtension:   File extension for the reassembled XML.
                   (default: `.xml`)
- postPurge:       Delete the disassembled directory after reassembly.
*/
import { ReassembleXMLFileHandler } from "xml-disassembler";

const handler = new ReassembleXMLFileHandler();
await handler.reassemble({
  filePath: "test/baselines/general/HR_Admin",
  fileExtension: "permissionset-meta.xml",
  postPurge: true,
});
```

## Example

**Input XML file (`HR_Admin.permissionset-meta.xml`)**

```xml
<?xml version="1.0" encoding="UTF-8"?>
<PermissionSet xmlns="http://soap.sforce.com/2006/04/metadata">
    <applicationVisibilities>
        <application>JobApps__Recruiting</application>
        <visible>true</visible>
    </applicationVisibilities>
    <classAccesses>
        <apexClass>Send_Email_Confirmation</apexClass>
        <enabled>true</enabled>
    </classAccesses>
    <fieldPermissions>
        <editable>true</editable>
        <field>Job_Request__c.Salary__c</field>
        <readable>true</readable>
    </fieldPermissions>
    <description>Grants all rights needed for an HR administrator to manage employees.</description>
    <label>HR Administration</label>
    <userLicense>Salesforce</userLicense>
    <objectPermissions>
        <allowCreate>true</allowCreate>
        <allowDelete>true</allowDelete>
        <allowEdit>true</allowEdit>
        <allowRead>true</allowRead>
        <viewAllRecords>true</viewAllRecords>
        <modifyAllRecords>true</modifyAllRecords>
        <object>Job_Request__c</object>
    </objectPermissions>
    <pageAccesses>
        <apexPage>Job_Request_Web_Form</apexPage>
        <enabled>true</enabled>
    </pageAccesses>
    <recordTypeVisibilities>
        <recordType>Recruiting.DevManager</recordType>
        <visible>true</visible>
    </recordTypeVisibilities>
    <tabSettings>
        <tab>Job_Request__c</tab>
        <visibility>Available</visibility>
    </tabSettings>
    <userPermissions>
        <enabled>true</enabled>
        <name>APIEnabled</name>
    </userPermissions>
</PermissionSet>
```

**Disassembled XML Directory**

<img src="https://raw.githubusercontent.com/mcarvin8/xml-disassembler/main/.github/images/disassembled.png">
<p><em>Disassembled XML files using unique ID elements</em></p>
<br>

<img src="https://raw.githubusercontent.com/mcarvin8/xml-disassembler/main/.github/images/disassembled-hashes.png">
<p><em>Disassembled XML files using SHA-256 hashes</em></p>

## Use Case

See [`sf-decomposer`](https://github.com/mcarvin8/sf-decomposer) for a Salesforce CLI use case:

- [Disassemble Use Case](https://github.com/mcarvin8/sf-decomposer/blob/main/src/service/decomposeFileHandler.ts)
- [Reassemble Use Case](https://github.com/mcarvin8/sf-decomposer/blob/main/src/service/recomposeFileHandler.ts)

## Ignore File

Create an ignore file (`.xmldisassemblerignore` by default) to exclude XMLs from disassembly. 

- uses [`node-ignore`](https://github.com/kaelzhang/node-ignore) (follows [.gitignore spec 2.22.1](https://git-scm.com/docs/gitignore))

## XML Parser

Uses [`fast-xml-parser`](https://github.com/NaturalIntelligence/fast-xml-parser) with support for:

- Character Data (CDATA): `"![CDATA["`
- Comments: `"!---"`
- Attributes: `"@__**"`

## Logging

Instead of printing to the terminal, the disassembler uses [`log4js`](https://github.com/log4js-node/log4js-node) to create a logging file. Logs are stored in `disassemble.log`. 

By default, only errors are logged.

```
[2024-03-30T14:28:37.950] [ERROR] default - The XML file HR_Admin.no-nested-elements.xml only has leaf elements. This file will not be disassembled.
```

Enable debug logs using the `setLogLevel` function:

```typescript
import {
  DisassembleXMLFileHandler,
  ReassembleXMLFileHandler,
  setLogLevel,
} from "xml-disassembler";

setLogLevel("debug");
```

## Extensions

These extensions expand on `xml-disassembler`:

- [`xml2json-disassembler`](https://github.com/mcarvin8/xml2json-disassembler): Disassemble large XML files into smaller JSON files and reassemble the original XML file when needed
- [`xml2yaml-disassembler`](https://github.com/mcarvin8/xml2yaml-disassembler): Disassemble large XML files into smaller YAML files and reassemble the original XML file when needed
- [`xml2json5-disassembler`](https://github.com/mcarvin8/xml2json5-disassembler): Disassemble large XML files into smaller JSON5 files and reassemble the original XML file when needed

## Contributing

Contributions are welcome! See [Contributing](https://github.com/mcarvin8/xml-disassembler/blob/main/CONTRIBUTING.md).

## Template

This project was created from a template by [Allan Oricil](https://github.com/AllanOricil).

His original [license](https://github.com/AllanOricil/js-template/blob/main/LICENSE) remains in this project.
